searchNodes=[{"doc":"Carbonite makes it easy to create audit trails for tables in a PostgreSQL database and integrate them into an Elixir application. Carbonite implements the Change-Data-Capture pattern on top of a PostgreSQL database. It uses triggers to automatically record all changes applied to a database table in order to guarantee a complete audit trail of the contained data. It is centered around the idea that the database transaction is the natural auditing unit of data mutation. Any mutation on a participating table requires the developer first to record the operation metadata within the same database transaction. The metadata record is associated to the table change records by a constraint. On top of its database layer, Carbonite provides an API allowing developers to effortlessly retrieve, process, and purge the audit trails within the Elixir application. How it works Carbonite keeps a central changes table where all mutations of participating tables are recorded. On each such table a trigger is installed (after INSERT , UPDATE , and DELETE statements - TRUNCATE is not supported, see below) which calls a procedure stored within the database. This procedure captures the new or updated data within the changes table automatically. The procedure fetches its own per-table settings from another table called triggers . These settings customize the procedure's behaviour, for instance, they may exclude certain table columns from being captured. Besides the changes table, storing information about individual database statements (or rather, their impact on the data), Carbonite's transactions table stores information on database transactions . Within your application logic, you can insert a row into this table after you begin a database transaction, and record arbitrary metadata for it. Apart from the metadata, the transactions table houses two identifier columns: id and xact_id . While id is an ordinary, autoincrementing integer primary key, the xact_id is set to pg_current_xact_id() , PostgreSQL's internal transaction identifier. Each row in the changes table is associated to a single row in the transactions table, referencing both these identifiers . The changes.transaction_id (referencing transactions.id ) field is filled with the &quot;current value&quot; of the related sequence. This is your regular foreign key column, enough to relate a changes record with exactly one transactions record. Consistency with the transactions table is guaranteed by a foreign key constraint. Additionally, the changes.transaction_xact_id field is set to pg_current_xact_id() . Together these references ensure that, not only relates each changes record to a single transactions record, but records in these tables have been inserted within the same database transaction . Consistency is ensure by a manual lookup in the trigger procedure. This leads to the following interesting properties: All changes created within a database transaction automatically and implicitly belong to the same record in the transactions table, even if they're created separately and agnostic of each other in the application logic. This gives the developer a &quot;natural&quot; way to group related changes into events (more on events later). The entry in the transactions table must be created before any changes . Attempting to modify an audited table without prior insertion into the transactions table will result in an error. ℹ️ &ensp;Trigger vs. Write-Ahead-Log / Logical Decoding / Extensions Existing solutions for CDC on top of a PostgreSQL database (e.g. Debezium ) often tail the Write-Ahead-Log or use extensions (e.g. pgaudit ) instead of using triggers to create change records. Both approaches have their own advantages and disadvantages, and one should try to understand the differences before making up their mind. One of the main selling points of triggers is that they are executed as part of the normal transaction logic in the database, hence they benefit from the usual atomicity and consistency guarantees of relational databases. For audit triggers, among other things this means that data mutations and their audit trail are committed as an atomic unit. At the same time, this property of triggers strongly couples the auditing logic to the business operation. For instance, if the audit trigger raises an error, the entire transaction is aborted. And vice versa, if the business operation aborts, nothing is audited. Additionally, auditing unavoidably has an impact on the database performance. A common criticism of audit triggers is the performance penalty they incur on otherwise &quot;simple&quot; database operations, which is said to be much higher than just tailing the WAL. Carbonite tries to limit the work done in its trigger, but a few indexes and tables are nonetheless touched. Another advantage of triggers is their universality: You should be able to run Carbonite's migrations on any hosted PostgreSQL instance without the need to tweak its configuration or install custom extensions before. The primary advantage of triggers for Carbonite, though, is that we can make use of PostgreSQL's internal transaction id: It allows us to group related changes together and record metadata for the group. This and the simple tooling around it is not (as easily) achievable with a fully decoupled system. However, if you do not need this feature of Carbonite, please be sure to consider other solutions as well. ℹ️ &ensp;Why TRUNCATE is not supported As the last of the 4 primary data mutating SQL commands, TRUNCATE allows to delete all data from a table or a set of tables. It can be instrumented using triggers on the statement level ( FOR EACH STATEMENT instead of FOR EACH ROW ), which means the trigger procedure executes only once for the statement and without any data - in contrast to, for instance, an UPDATE statement, which also mutates multiple rows but fires the procedure once for each row. While there might be value in auditing TRUNCATE statements, the behaviour of the trigger procedure would be quite different from the other commands. For a rarely used SQL command, we chose against this additional complexity and not to support TRUNCATE in Carbonite. Installation Requirements Due to its use of pg_current_xact_id , Carbonite requires PostgreSQL version 13 or above . If you see an error message like the following, your PostgreSQL installation is too old: ** (Postgrex.Error) ERROR 42704 (undefined_object) type &quot;xid8&quot; does not exist Hex dependency # mix.exs def deps do [ { :carbonite , &quot;~&gt; 0.6.0&quot; } ] end Getting started As much of Carbonite's logic lives in database functions and triggers, to get started, we first need to create a migration using Ecto. Creating the initial migration Carbonite contains a Mix task that generates the initial migration for you. Please open the generated file and edit it according to your needs. See Carbonite.Migrations for more information on migrations. mix carbonite.gen.initial_migration -r MyApp.Repo The final migration should look something like this: # priv/repo/migrations/20210704201534_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do use Ecto.Migration def up do Carbonite.Migrations . up ( 1 ) Carbonite.Migrations . up ( 2 ) Carbonite.Migrations . up ( 3 ) Carbonite.Migrations . up ( 4 ) Carbonite.Migrations . up ( 5 ) # For each table that you want to capture changes of, you need to install the trigger. Carbonite.Migrations . create_trigger ( :rabbits ) # Optionally you may configure the trigger inside the migration. Carbonite.Migrations . put_trigger_config ( :rabbits , :excluded_columns , [ &quot;age&quot; ] ) # Optional outbox to process transactions later. Carbonite.Migrations . create_outbox ( &quot;rabbit_holes&quot; ) end def down do # Remove outbox again. Carbonite.Migrations . drop_outbox ( &quot;rabbit_holes&quot; ) # Remove all triggers before dropping the schema. Carbonite.Migrations . drop_trigger ( :rabbits ) # Drop the Carbonite tables. Carbonite.Migrations . down ( 5 ) Carbonite.Migrations . down ( 4 ) Carbonite.Migrations . down ( 3 ) Carbonite.Migrations . down ( 2 ) Carbonite.Migrations . down ( 1 ) end end Updates When a new Carbonite version is released, it may contain updates to the database schema. As these are announced in the Changelog , you each time need to create a migration in your host application like the following. # priv/repo/migrations/20210704201534_update_carbonite.exs defmodule MyApp.Repo.Migrations.UpdateCarbonite do use Ecto.Migration def up do Carbonite.Migrations . up ( 3 ) end def down do Carbonite.Migrations . down ( 3 ) end end Applying multiple Carbonite migrations in a single host migration is fine. Note that for each of your Carbonite &quot;partitions&quot; (see below), you need to run each Carbonite migration. Trigger configuration The behaviour of the capture trigger is customizable per table by manipulating the settings in the triggers table. Often you will want to update these settings within a migration, as well, which is why Carbonite provides the small helper function Carbonite.Migrations.put_trigger_config/4 that updates the settings using plain SQL statements. Primary Key Columns To speed up version lookups for a specific record, Carbonite copies its primary key(s) to the table_pk column of the changes table. The table keeps an index on this column together with the table prefix and name. By default, Carbonite will try to copy the :id column of the source table. If your table does not have a primary key, has a primary key with a different name, or has a composite primary key, you can override this using the primary_key_columns option. # Disable PK copying Carbonite.Migrations . put_trigger_config ( :rabbits , :primary_key_columns , [ ] ) # Different name Carbonite.Migrations . put_trigger_config ( :rabbits , :primary_key_columns , [ &quot;identifier&quot; ] ) # Composite PK Carbonite.Migrations . put_trigger_config ( :rabbits , :primary_key_columns , [ &quot;house&quot; , &quot;apartment_no&quot; ] ) Since the changes table keeps versions of a multitude of different source tables, primary keys are first cast to string (the table_pk column has type VARCHAR[] ). For composite primary keys, set the primary_key_columns option to an array as shown above. Each component of a compound primary key will be cast to string before the components are joined into the array. Excluded and Filtered Columns In case your table contains sensitive data or data otherwise undesirable for change capturing, you can exclude columns using the excluded_columns option. Excluded columns will not appear in the captured data. If an UPDATE on a table solely touches excluded columns, the entire UPDATE will not be recorded. Carbonite.Migrations . put_trigger_config ( :rabbits , :excluded_columns , [ &quot;age&quot; ] ) If you still want to capture changes to a column (in the changed field), but don't need the exact data, you can make it a &quot;filtered&quot; column. These columns appear as [FILTERED] in the data field. Carbonite.Migrations . put_trigger_config ( :rabbits , :filtered_columns , [ &quot;age&quot; ] ) Persisting replaced data on UPDATE statements By default, the changes table will track the current version of a record in its data field and a list of changed fields in the changed column. Optionally, you can decide to store the replaced data (the &quot;diff&quot;) in the changed_from field, by enabling the store_changed_from option for a trigger. This defaults to false . Carbonite.Migrations . put_trigger_config ( :rabbits , :store_changed_from , true ) Partitioning the Audit Trail Carbonite can install its tables into multiple database schemas using the carbonite_prefix option. You can use this feature to &quot;partition&quot; your captured data. Carbonite.Migrations . up ( 1 , carbonite_prefix : &quot;carbonite_animals&quot; ) Carbonite.Migrations . create_trigger ( :rabbits , carbonite_prefix : &quot;carbonite_animals&quot; ) Basically all of Carbonite's functions accept the same carbonite_prefix option to target a particular partition. If desired, tables can participate in multiple partitions by adding multiple triggers on to them. Keep in mind that each partition will need to be processed and purged separately, resulting in multiple streams of change data in your external storage. Inserting a Transaction In your application logic, before modifying an audited table like rabbits , you need to first create a Carbonite.Transaction record. With Ecto.Multi The easiest way to do so is using Carbonite.Multi.insert_transaction/2 within an Ecto.Multi operation: Ecto.Multi . new ( ) |&gt; Carbonite.Multi . insert_transaction ( %{ meta : %{ type : &quot;rabbit_inserted&quot; } } ) |&gt; Ecto.Multi . insert ( :rabbit , &amp; MyApp.Rabbit . create_changeset ( &amp;1 . params ) ) |&gt; MyApp.Repo . transaction ( ) As you can see, the Carbonite.Transaction is a great place to store metadata for the operation. A type field can be used to categorize the transactions. A user_id would be a good candidate for a transaction log, as well. Building a changeset for manual insertion If you don't have the luxury of an Ecto.Multi , you can create a changeset for a Carbonite.Transaction using Carbonite.Transaction.changeset/1 : MyApp.Repo . transaction ( fn -&gt; %{ meta : %{ type : &quot;rabbit_inserted&quot; } } |&gt; Carbonite.Transaction . changeset ( ) |&gt; MyApp.Repo . insert! ( ) # ... end ) Setting metadata outside of the transaction In case you do not have access to metadata you want to persist in the Carbonite.Transaction at the code site where you create it, you can use Carbonite.Transaction.put_meta/2 to store metadata in the process dictionary . This metadata is merged into the metadata given to Carbonite.Multi.insert_transaction/2 . # e.g., in a controller or plug Carbonite.Transaction . put_meta ( :user_id , ... ) Transactions in data migrations If you manipulate data inside your transactions, as usual a Carbonite.Transaction needs to be inserted before any other statements. You can use Carbonite.Migrations.insert_migration_transaction/1 to insert a transaction with a meta attribute populated from the migration module. import Carbonite.Migrations def change do insert_migration_transaction ( ) execute ( &quot;UPDATE ...&quot; ) end Retrieving data Of course, persisting the audit trail is not an end in itself. At some point you will want to read the data back and make it accessible to the user. Carbonite.Query offers a small suite of helper functions that make it easier to query the database for Transaction and Change records. Fetching transactions The Carbonite.Query.transactions/1 function constructs an Ecto.Query for loading Carbonite.Transaction.t/0 records from the database, optionally preloading their included changes. The query can be further refined to limit the result set. Carbonite.Query . transactions ( ) |&gt; Ecto.Query . where ( [ t ] , t . inserted_at &gt; ^ earliest ) |&gt; MyApp.Repo . all ( ) Fetching changes of invididual records The Carbonite.Query.changes/2 function constructs an Ecto.Query.t/0 from a schema struct, loading all changes stored for the given source record. % MyApp.Rabbit { id : 1 } |&gt; Carbonite.Query . changes ( ) |&gt; MyApp.Repo . all ( ) Processing data Storing event series in a relational database comes with the usual scaling issues, so at some point you may find it necessary to offload the captured data to an external data store. For processing, exporting, and later purging the captured transactions, Carbonite includes what is called an &quot;Event Outbox&quot;, or in fact as many of those as you need. Essentially, a Carbonite.Outbox is a named cursor on the ordered list of transactions stored in an audit trail. Each outbox advances by calling a user-supplied processor function on a batch of transactions. The batch can be filtered and limited in size, but is always ordered ascending by the transaction's id attribute, so transactions are processed roughly in order of their insertion (see below for a discussion on caveats of this solution). When a batch is processed, the outbox remembers its last position - the last transaction that has been processed - so the following call to the outbox can continue where the last has left of. Besides, the outbox persists a user-definable &quot;memo&quot;, an arbitrary map that allows to feed output of the previous processor run into the next one. Creating an Outbox Create an outbox within a migration: Carbonite.Migrations . create_outbox ( &quot;rabbit_holes&quot; ) You can create more than one outbox if you need multiple processors, e.g. for multiple target systems for your data. Processing You can process an outbox by calling Carbonite.process/4 with the outbox name and a processor callback function. Carbonite . process ( MyApp.Repo , &quot;rabbit_holes&quot; , fn transactions , _memo -&gt; for transaction &lt;- transactions do send_to_external_database ( transaction ) end :cont end ) In practise, you will almost always want to run this within an asynchronous job processor. The following shows an exemplary Oban worker: # config/config.exs config :my_app , :oban , repo : MyApp.Repo , plugins : [ { Oban.Plugins.Cron , crontab : [ { &quot;0 0 * * *&quot; , MyApp.PeriodicOutboxWorker , args : %{ outbox : &quot;rabbit_holes&quot; } } ] ] # lib/my_app/periodic_outbox_worker.ex defmodule MyApp.PeriodicOutboxWorker do # Worker is scheduled every 24h, let&#39;s make it unique for 12h to be sure. # Retry is disabled, the next scheduled run will pick up failed outboxes. use Oban.Worker , queue : :default , unique : [ period : 43_200 ] , retry : false def perform ( % Oban.Job { args : %{ &quot;outbox&quot; =&gt; outbox } } ) do Carbonite . process ( MyApp.Repo , outbox , fn transactions , _memo -&gt; for transaction &lt;- transactions do send_to_external_database ( transaction ) end :cont end ) end end ⚠️&ensp;Transactionality & mutual exclusion As the outbox processor is likely to call external services as part of its work, Carbonite.process/4 does not begin a transaction itself, and consequently does not acquire a lock on the outbox record. In other words, users have to ensure that they don't accidentally run Carbonite.process/4 for the same outbox concurrently, for instance making use of uniqueness options of the used job processor (e.g. unique in Oban ). The absence of transactionality also means that any exception raised within the processor function will immediately abort the Carbonite.process/4 call without writing the current outbox position to disk. As a result, transactions may be processed again in the next run. Please make sure your receiving system or database can handle duplicate messages. ⚠️&ensp;Long running / parallel transactions and the outbox order When a Carbonite.Transaction record is created at the beginning of an operation in your application, it records the &quot;current&quot; sequence value in its id field. The record in the transactions table becomes visible to other transactions when the current transaction is committed. A few observations can be made from this: When ordered by their id field, the transactions will be roughly sorted by the time the corresponding operation was started, not when it was committed. Two transactions running in parallel may be committed &quot;out of order&quot;, i.e. one with the larger id may be committed before the smaller id transaction if that has a longer runtime. The latter point is crucial: For instance, if two transactions with id=1 and id=2 run in parallel and id=2 finishes before id=1 , an outside viewer can already see id=2 in the database before id=1 is committed. If this outside viewer happens to be an outbox processing job, transaction with id=1 might be skipped and never looked at again. To mitigate this issue, Carbonite.process/4 has a min_age option which excludes transaction younger than a certain from the processing batch (5 minutes by default, increase this is you expect longer running transactions). Purging Now that you have successfully exported your audit data, you may delete old transactions from the primary database. Carbonite.purge/2 deletes records that have been processed by all existing outboxes. # Deletes records whose id field is less than the transaction_id on each outbox. Carbonite . purge ( MyApp.Repo ) Testing / Bypassing Carbonite One of Carbonite's key features is that it is virtually impossible to forget to record a change to a table (due to the trigger) or to forget to insert an enclosing Carbonite.Transaction beforehand (due to the foreign key constraint between changes and transactions ). However, in some circumstances it may be desirable to temporarily switch off change capturing. One such situation is the use of factories (e.g. ExMachina) inside your test suite: Inserting a transaction before each factory call quickly becomes cumbersome and will unnecessarily increase execution time. To bypass the capture trigger, Carbonite's trigger configuration provides a toggle mechanism consisting of two fields: mode and override_xact_id . The former you set while installing the trigger on a table in a migration, while the latter allows to &quot;override&quot; whatever has been set at runtime, and only for the current transaction. If you are using Ecto's SQL sandbox for running transactional tests, this means the override is going to be active until the end of the test case. As a result, you have two options: Leave the mode at the default value of :capture and turn off capturing as needed by switching to &quot;override mode&quot;. This means for every test case where you do not care about change capturing, you explicitly disable the trigger before any database calls; for instance, in an ExUnit setup block. This approach has the benefit that you still capture all changes by default, and can't miss to test a code path that (in production) would require a Carbonite.Transaction . It is, however, still pretty expensive at ~1 additional SQL call per test case. Set the mode to :ignore on all triggers in your :test environment and instead selectively turn on capturing in test cases where you want to assert on the captured data. For instance, you can set the trigger mode in your migration based on the Mix environment. This approach is cheaper as it does not require any action in your tests by default. Yet you should make sure that you test all code paths that do mutate change-captured tables, in order to assert that each of these inserts a transaction as well. The following code snippet illustrates the second approach: # config/config.exs config :my_app , carbonite_mode : :capture # config/test.exs config :my_app , carbonite_mode : :ignore # priv/repo/migrations/000000000000_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do @mode Application . compile_env! ( :my_app , :carbonite_mode ) def up do # ... Carbonite.Migrations . create_trigger ( :rabbits ) Carbonite.Migrations . put_trigger_config ( :rabbits , :mode , @mode ) end end # test/support/carbonite_helpers.exs defmodule MyApp.CarboniteHelpers do def carbonite_override_mode ( _ ) do Carbonite . override_mode ( MyApp.Repo ) :ok end def current_transaction_meta do Carbonite.Query . current_transaction ( ) |&gt; MyApp.Repo . one! ( ) |&gt; Map . fetch ( :meta ) end end # test/some_test.exs describe &quot;my_operation/0&quot; do setup [ :carbonite_override_mode ] test &quot;auditing&quot; do my_operation ( ) assert current_transaction_meta ( ) == %{ &quot;type&quot; =&gt; &quot;some_operation&quot; } end end","ref":"Carbonite.html","title":"Carbonite","type":"module"},{"doc":"Returns the default audit trail prefix.","ref":"Carbonite.html#default_prefix/0","title":"Carbonite.default_prefix/0","type":"function"},{"doc":"Fetches all changes of the current transaction from the database. Make sure to run this within a transaction. Parameters repo - the Ecto repository opts - optional keyword list Options carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.html#fetch_changes/2","title":"Carbonite.fetch_changes/2","type":"function"},{"doc":"Inserts a Carbonite.Transaction.t/0 into the database. Make sure to run this within a transaction. Parameters repo - the Ecto repository params - map of params for the Carbonite.Transaction (e.g., :meta ) opts - optional keyword list Options carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot; Multiple inserts in the same transaction Normally, you should have exactly one insert_transaction/3 call per database transaction. In practise, there are two scenarios in this function may be called multiple times: If an operation A, which calls insert_transaction/3 , sometimes is nested within an outer operation B, which also calls insert_transaction/3 . In tests using Ecto's SQL sandbox, subsequent calls to transactional operations (even to the same operation twice) are wrapped inside the overarching test transaction, and hence also effectively call insert_transaction/3 within the same transaction. While the first scenario can be resolved using appropriate control flow (e.g. by conditionally disabling the inner insert_transaction/3 call), the second scenario is quite common and often unavoidable. Therefore, insert_transaction/3 ignores subsequent calls within the same database transaction (equivalent to ON CONFLICT DO NOTHING ), discarding metadata passed to all calls but the first.","ref":"Carbonite.html#insert_transaction/3","title":"Carbonite.insert_transaction/3","type":"function"},{"doc":"Sets the current transaction to &quot;override mode&quot; for all tables in the audit log. Parameters repo - the Ecto repository opts - optional keyword list Options carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.html#override_mode/2","title":"Carbonite.override_mode/2","type":"function"},{"doc":"Processes an outbox queue. This function sends chunks of persisted transactions to a user-supplied processing function. It looks up the current &quot;reading position&quot; from a given Carbonite.Outbox and yields transactions matching the given filter criteria ( min_age , etc.) until either the input source is exhausted or a processing function application returns :halt . Returns the {:ok, outbox} or {:halt, outbox} depending on whether processing was halted explicitly or due the exhausted input source. See Carbonite.Query.outbox_queue/2 for query options. Examples Carbonite . process ( MyApp.Repo , &quot;rabbit_holes&quot; , fn [ transaction ] , _memo -&gt; # The transaction has its changes preloaded. transaction |&gt; MyApp.Foo . serialize ( ) |&gt; MyApp.Foo . send_to_external_database ( ) :cont end ) Memo passing The memo is useful to carry data between each processor application. Let's say you wanted to generate a hashsum chain on your processed data: Carbonite . process ( MyApp.Repo , &quot;rabbit_holes&quot; , fn [ transaction ] , %{ &quot;checksum&quot; =&gt; checksum } -&gt; { payload , checksum } = MyApp.Foo . serialize_and_hash ( transaction , checksum ) MyApp.Foo . send_to_external_database ( payload ) { :cont , memo : %{ &quot;checksum&quot; =&gt; checksum } } end ) Chunking / Limiting The examples above received a single-element list as their first parameter: This is because the transactions are actually processed in &quot;chunks&quot; and the default chunk size is 1. If you would like to process more transactions in one chunk, set the chunk option: Carbonite . process ( MyApp.Repo , &quot;rabbit_holes&quot; , [ chunk : 50 ] , fn transactions , _memo -&gt; for transaction &lt;- transactions do transaction |&gt; MyApp.Foo . serialize ( ) |&gt; MyApp.Foo . send_to_external_database ( ) end :cont end ) The query that is executed to fetch the data from the database is controlled with the limit option and is independent of the chunk size. Error handling In case you run into an error midway into processing a batch, you may choose to halt processing while remembering about the last processed transaction. Carbonite . process ( MyApp.Repo , &quot;rabbit_holes&quot; , fn [ transaction ] , _memo -&gt; case send_to_external_database ( transaction ) do :ok -&gt; :cont { :error , _term } -&gt; :halt end end You can, however, if you know the first half of a batch has been processed, still update the memo and last_transaction_id . Carbonite . process ( MyApp.Repo , &quot;rabbit_holes&quot; , fn transactions , _memo -&gt; case process_transactions ( transactions ) do { :error , last_successful_transaction } -&gt; { :halt , last_transaction_id : last_successful_transaction . id } :ok -&gt; :cont end end Parameters repo - the Ecto repository outbox_name - name of the outbox to process opts - optional keyword list process_func - see process_func/0 for details Options min_age - the minimum age of a record, defaults to 300 seconds (set nil to disable) limit - limits the query in size, defaults to 100 (set nil to disable) filter - function for refining the batch query, defaults to nil chunk - defines the size of the chunk passed to the process function, defaults to 1 carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.html#process/4","title":"Carbonite.process/4","type":"function"},{"doc":"Deletes transactions that have been fully processed. See Carbonite.Query.outbox_done/1 for query options. Returns the number of deleted transactions. Parameters repo - the Ecto repository opts - optional keyword list Options min_age - the minimum age of a record, defaults to 300 seconds (set nil to disable) carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.html#purge/2","title":"Carbonite.purge/2","type":"function"},{"doc":"","ref":"Carbonite.html#t:prefix/0","title":"Carbonite.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.html#t:prefix_option/0","title":"Carbonite.prefix_option/0","type":"type"},{"doc":"This type defines the callback function signature for Carbonite.process/3 . The processor function receives the current chunk of transactions and the memo of the last function application, and must return one of :cont - continue processing :halt - stop processing after this chunk {:cont | :halt, opts} - cont/halt and set some options After the process function invocation the Outbox is updated with new attributes. Options Returned options can be: memo - memo to store on Outbox, defaults to previous memo last_transaction_id - last transaction id to remember as processed, defaults to previous ` last_transaction_id ` on ` :halt ` , defaults to last id in current chunk when ` :cont ` is returned","ref":"Carbonite.html#t:process_func/0","title":"Carbonite.process_func/0","type":"type"},{"doc":"","ref":"Carbonite.html#t:process_func_option/0","title":"Carbonite.process_func_option/0","type":"type"},{"doc":"","ref":"Carbonite.html#t:process_option/0","title":"Carbonite.process_option/0","type":"type"},{"doc":"","ref":"Carbonite.html#t:purge_option/0","title":"Carbonite.purge_option/0","type":"type"},{"doc":"","ref":"Carbonite.html#t:repo/0","title":"Carbonite.repo/0","type":"type"},{"doc":"A Carbonite.Change records a mutation on a database table. INSERT statements lead to a Change where the data field contains the inserted row as a JSON object while the changed field is an empty list. UPDATE statements contain the updated record in data while the changed field is a list of attributes that have changed. The changed_from field may additionally contain a partial representation of the replaced record, showing only the diff to the updated record. This behaviour is optional and can be enabled with the store_changed_from trigger option. DELETE statements have the deleted data in data while changed is again an empty list.","ref":"Carbonite.Change.html","title":"Carbonite.Change","type":"module"},{"doc":"","ref":"Carbonite.Change.html#t:id/0","title":"Carbonite.Change.id/0","type":"type"},{"doc":"","ref":"Carbonite.Change.html#t:t/0","title":"Carbonite.Change.t/0","type":"type"},{"doc":"Functions to setup Carbonite audit trails in your migrations.","ref":"Carbonite.Migrations.html","title":"Carbonite.Migrations","type":"module"},{"doc":"Inserts an outbox record into the database. Options carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#create_outbox/2","title":"Carbonite.Migrations.create_outbox/2","type":"function"},{"doc":"Installs a change capture trigger on a table. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#create_trigger/2","title":"Carbonite.Migrations.create_trigger/2","type":"function"},{"doc":"Rollback a migration. Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; drop_schema controls whether the initial migration deletes the schema during rollback","ref":"Carbonite.Migrations.html#down/2","title":"Carbonite.Migrations.down/2","type":"function"},{"doc":"Removes an outbox record. Options carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#drop_outbox/2","title":"Carbonite.Migrations.drop_outbox/2","type":"function"},{"doc":"Removes a change capture trigger from a table. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#drop_trigger/2","title":"Carbonite.Migrations.drop_trigger/2","type":"function"},{"doc":"Inserts a transaction for a data migration. The transaction's meta attribute is populated with { &quot;type&quot; : &quot;migration&quot; , direction : &quot;up&quot; } ... and additionally the name from the parameters. Example defmodule MyApp.Repo.Migrations.SomeDataMigration do use Ecto.Migration import Carbonite.Migrations def change do insert_migration_transaction ( ) execute ( &quot;UPDATE ...&quot; , &quot;...&quot; ) end end This works the same for up/0 / down/0 -style migrations: defmodule MyApp.Repo.Migrations.SomeDataMigration do use Ecto.Migration import Carbonite.Migrations def up do insert_migration_transaction ( &quot;some-data-migrated&quot; ) execute ( &quot;INSERT ...&quot; ) end def down do insert_migration_transaction ( &quot;some-data-rolled-back&quot; ) execute ( &quot;DELETE ...&quot; ) end end Options carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot; meta is a map with additional meta information to store on the transaction","ref":"Carbonite.Migrations.html#insert_migration_transaction/2","title":"Carbonite.Migrations.insert_migration_transaction/2","type":"function"},{"doc":"Defines a Ecto.Migration.after_begin/0 implementation for a data migration. See insert_migration_transaction/1 for options. This determines the name of the transaction from the migration's module name. { &quot;direction&quot; : &quot;up&quot; , &quot;name&quot; : &quot;my_app/repo/migrations/example&quot; , &quot;type&quot; : &quot;migration&quot; } Example defmodule MyApp.Repo.Migrations.SomeDataMigration do use Ecto.Migration import Carbonite.Migrations insert_migration_transaction_after_begin ( ) def change do execute ( &quot;UPDATE ...&quot; ) end end Alternatively, if you define your own migration template module: defmodule MyApp.Migration do defmacro __using__ do quote do use Ecto.Migration import Carbonite.Migrations insert_migration_transaction_after_begin ( ) end end end","ref":"Carbonite.Migrations.html#insert_migration_transaction_after_begin/1","title":"Carbonite.Migrations.insert_migration_transaction_after_begin/1","type":"macro"},{"doc":"Allows to update a trigger configuration option for a given table. This function builds an SQL UPDATE statement that can be used within a database migration to update a setting stored in Carbonite's triggers table without using the Carbonite.Trigger schema or other application-level code that is prone to change over time. This helps to ensure that your data migrations continue to function, regardless of future updates to Carbonite. Configuration values primary_key_columns is a list of columns that form the primary key of the table ( defaults to ` [ &quot;id&quot; ] ` , set to ` [ ] ` to disable ) excluded_columns is a list of columns to exclude from change captures filtered_columns is a list of columns that appear as '[FILTERED]' in the data store_changed_from is a boolean defining whether the changed_from field should be filled mode is either :capture or :ignore and defines the default behaviour of the trigger Example Carbonite.Migrations . put_trigger_config ( &quot;rabbits&quot; , :excluded_columns , [ &quot;name&quot; ] ) Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#put_trigger_config/4","title":"Carbonite.Migrations.put_trigger_config/4","type":"function"},{"doc":"Runs one of Carbonite's migrations. Migration patchlevels Make sure that you run all migrations in your host application. Initial patch: 1 Current patch: 5 Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#up/2","title":"Carbonite.Migrations.up/2","type":"function"},{"doc":"","ref":"Carbonite.Migrations.html#t:down_option/0","title":"Carbonite.Migrations.down_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:insert_migration_transaction_option/0","title":"Carbonite.Migrations.insert_migration_transaction_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:outbox_name/0","title":"Carbonite.Migrations.outbox_name/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:outbox_option/0","title":"Carbonite.Migrations.outbox_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:patch/0","title":"Carbonite.Migrations.patch/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:prefix/0","title":"Carbonite.Migrations.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:table_name/0","title":"Carbonite.Migrations.table_name/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:trigger_config_key/0","title":"Carbonite.Migrations.trigger_config_key/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:trigger_option/0","title":"Carbonite.Migrations.trigger_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:up_option/0","title":"Carbonite.Migrations.up_option/0","type":"type"},{"doc":"This module provides functions for dealing with audit trails in the context of Ecto.Multi.","ref":"Carbonite.Multi.html","title":"Carbonite.Multi","type":"module"},{"doc":"Adds a operation to an Ecto.Multi to fetch the changes of the current transaction. Useful for returning all transaction changes to the caller. Multi step is called :carbonite_changes . See Carbonite.fetch_changes/2 for options. Example Ecto.Multi . new ( ) |&gt; Carbonite.Multi . insert_transaction ( %{ meta : %{ type : &quot;create_rabbit&quot; } } ) |&gt; Ecto.Multi . insert ( :rabbit , fn _ -&gt; Rabbit . changeset ( %{ } ) end ) |&gt; Carbonite.Multi . fetch_changes ( )","ref":"Carbonite.Multi.html#fetch_changes/2","title":"Carbonite.Multi.fetch_changes/2","type":"function"},{"doc":"Adds an insert operation for a Carbonite.Transaction to an Ecto.Multi . Multi step is called :carbonite_transaction . See Carbonite.insert_transaction/3 for options. Example Ecto.Multi . new ( ) |&gt; Carbonite.Multi . insert_transaction ( %{ meta : %{ type : &quot;create_rabbit&quot; } } ) |&gt; Ecto.Multi . insert ( :rabbit , fn _ -&gt; Rabbit . changeset ( %{ } ) end )","ref":"Carbonite.Multi.html#insert_transaction/3","title":"Carbonite.Multi.insert_transaction/3","type":"function"},{"doc":"Sets the current transaction to &quot;override mode&quot; for all tables in the audit log. See Carbonite.override_mode/2 for options.","ref":"Carbonite.Multi.html#override_mode/2","title":"Carbonite.Multi.override_mode/2","type":"function"},{"doc":"","ref":"Carbonite.Multi.html#t:params/0","title":"Carbonite.Multi.params/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:prefix/0","title":"Carbonite.Multi.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:prefix_option/0","title":"Carbonite.Multi.prefix_option/0","type":"type"},{"doc":"A Carbonite.Outbox stores metadata for outboxes like the last processed transaction. The last_transaction_id field defaults to zero, indicating that nothing has been processed yet.","ref":"Carbonite.Outbox.html","title":"Carbonite.Outbox","type":"module"},{"doc":"Builds an update changeset.","ref":"Carbonite.Outbox.html#changeset/2","title":"Carbonite.Outbox.changeset/2","type":"function"},{"doc":"","ref":"Carbonite.Outbox.html#t:memo/0","title":"Carbonite.Outbox.memo/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:name/0","title":"Carbonite.Outbox.name/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:t/0","title":"Carbonite.Outbox.t/0","type":"type"},{"doc":"This module provides query functions for retrieving audit trails from the database.","ref":"Carbonite.Query.html","title":"Carbonite.Query","type":"module"},{"doc":"Returns an Ecto.Query.t/0 that can be used to select changes for a single record. Given an Ecto.Schema.t/0 struct, this function builds a query that fetches all changes recorded for it from the database, ordered ascending by their ID (i.e., roughly by insertion date descending). Example % MyApp.Rabbit { id : 1 } |&gt; Carbonite.Query . changes ( ) |&gt; MyApp.Repo . all ( ) Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; table_prefix allows to override the table prefix, defaults to schema prefix of the record preload can be used to preload the transaction","ref":"Carbonite.Query.html#changes/2","title":"Carbonite.Query.changes/2","type":"function"},{"doc":"Returns an Ecto.Query.t/0 that can be used to select or delete the &quot;current&quot; transaction. This function is useful when your tests run in a database transaction using Ecto's SQL sandbox. Example: Asserting on the current transaction When you insert your Carbonite.Transaction record somewhere inside your domain logic, you do not wish to return it to the caller only to be able to assert on its attributes in tests. This example shows how you could assert on the metadata inserted. # Test running inside Ecto&#39;s SQL sandbox. test &quot;my test&quot; do some_operation_with_a_transaction ( ) assert current_transaction_meta ( ) == %{ &quot;type&quot; =&gt; &quot;some_operation&quot; } end defp current_transaction_meta do Carbonite.Query . current_transaction ( ) |&gt; MyApp.Repo . one! ( ) |&gt; Map . fetch ( :meta ) end Options carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload - can be used to preload the changes, defaults to nil","ref":"Carbonite.Query.html#current_transaction/1","title":"Carbonite.Query.current_transaction/1","type":"function"},{"doc":"Returns an Ecto.Query.t/0 that selects a outbox by name. Options carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Query.html#outbox/2","title":"Carbonite.Query.outbox/2","type":"function"},{"doc":"Returns an Ecto.Query.t/0 that selects all completely processed transactions. If no outbox exists, this query returns all transactions. If one or more outboxes exist, this query returns all transactions with an ID less than the minimum of the last_transaction_id attributes of the outboxes. Transactions are not ordered. Options min_age - the minimum age of a record, defaults to 300 seconds (set nil to disable) carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload - can be used to preload the changes, defaults to true","ref":"Carbonite.Query.html#outbox_done/1","title":"Carbonite.Query.outbox_done/1","type":"function"},{"doc":"Returns an Ecto.Query.t/0 that selects the next batch of transactions for an outbox. Transactions are ordered by their ID ascending, so roughly in order of insertion. Options min_age - the minimum age of a record, defaults to 300 seconds (set nil to disable) limit - limits the query in size, defaults to 100 (set nil to disable) carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload - can be used to preload the changes, defaults to true","ref":"Carbonite.Query.html#outbox_queue/2","title":"Carbonite.Query.outbox_queue/2","type":"function"},{"doc":"Returns an Ecto.Query.t/0 that can be used to select transactions from the database. Examples Carbonite.Query . transactions ( ) |&gt; MyApp.Repo . all ( ) # Preload changes Carbonite.Query . transactions ( preload : :changes ) |&gt; MyApp.Repo . all ( ) # Same Carbonite.Query . transactions ( preload : true ) |&gt; MyApp.Repo . all ( ) Options carbonite_prefix - defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload - can be used to preload the changes, defaults to nil","ref":"Carbonite.Query.html#transactions/1","title":"Carbonite.Query.transactions/1","type":"function"},{"doc":"","ref":"Carbonite.Query.html#t:changes_option/0","title":"Carbonite.Query.changes_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:current_transaction_option/0","title":"Carbonite.Query.current_transaction_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:disabled/0","title":"Carbonite.Query.disabled/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:outbox_done_option/0","title":"Carbonite.Query.outbox_done_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:outbox_queue_option/0","title":"Carbonite.Query.outbox_queue_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:prefix/0","title":"Carbonite.Query.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:prefix_option/0","title":"Carbonite.Query.prefix_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:preload_option/0","title":"Carbonite.Query.preload_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:transactions_option/0","title":"Carbonite.Query.transactions_option/0","type":"type"},{"doc":"A Carbonite.Transaction is the binding link between change records of tables. As such, it contains a set of optional metadata that describes the transaction.","ref":"Carbonite.Transaction.html","title":"Carbonite.Transaction","type":"module"},{"doc":"Builds a changeset for a new Carbonite.Transaction . The :meta map from the params will be merged with the metadata currently stored in the process dictionary.","ref":"Carbonite.Transaction.html#changeset/1","title":"Carbonite.Transaction.changeset/1","type":"function"},{"doc":"Returns the currently stored metadata.","ref":"Carbonite.Transaction.html#current_meta/0","title":"Carbonite.Transaction.current_meta/0","type":"function"},{"doc":"Stores a piece of metadata in the process dictionary. This can be useful in situations where you want to record a value at a system boundary (say, the user's account_id ) without having to pass it through to the database transaction. Returns the currently stored metadata.","ref":"Carbonite.Transaction.html#put_meta/2","title":"Carbonite.Transaction.put_meta/2","type":"function"},{"doc":"","ref":"Carbonite.Transaction.html#t:id/0","title":"Carbonite.Transaction.id/0","type":"type"},{"doc":"","ref":"Carbonite.Transaction.html#t:meta/0","title":"Carbonite.Transaction.meta/0","type":"type"},{"doc":"","ref":"Carbonite.Transaction.html#t:t/0","title":"Carbonite.Transaction.t/0","type":"type"},{"doc":"A Carbonite.Trigger stores per table configuration for the change capture trigger.","ref":"Carbonite.Trigger.html","title":"Carbonite.Trigger","type":"module"},{"doc":"","ref":"Carbonite.Trigger.html#t:id/0","title":"Carbonite.Trigger.id/0","type":"type"},{"doc":"","ref":"Carbonite.Trigger.html#t:mode/0","title":"Carbonite.Trigger.mode/0","type":"type"},{"doc":"","ref":"Carbonite.Trigger.html#t:t/0","title":"Carbonite.Trigger.t/0","type":"type"},{"doc":"Generates a sample migration that runs all migrations currently contained within Carbonite.","ref":"Mix.Tasks.Carbonite.Gen.InitialMigration.html","title":"Mix.Tasks.Carbonite.Gen.InitialMigration","type":"task"},{"doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS APPENDIX: How to apply the Apache License to your work. To apply the Apache License to your work, attach the following boilerplate notice, with the fields enclosed by brackets &quot;[]&quot; replaced with your own identifying information. (Don't include the brackets!) The text should be enclosed in the appropriate comment syntax for the file format. We also recommend that a file or class name and description of purpose be included on the same &quot;printed page&quot; as the copyright notice for easier identification within third-party archives. Copyright 2021-2022 Bitcrowd GmbH Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"license.html","title":"LICENSE","type":"extras"},{"doc":"Carbonite makes it easy to create audit trails for tables in a PostgreSQL database and integrate them into an Elixir application. Carbonite implements the Change-Data-Capture pattern on top of a PostgreSQL database. It uses triggers to automatically record all changes applied to a database table in order to guarantee a complete audit trail of the contained data. It is centered around the idea that the database transaction is the natural auditing unit of data mutation. Any mutation on a participating table requires the developer first to record the operation metadata within the same database transaction. The metadata record is associated to the table change records by a constraint. On top of its database layer, Carbonite provides an API allowing developers to effortlessly retrieve, process, and purge the audit trails within the Elixir application.","ref":"readme.html","title":"README","type":"extras"},{"doc":"Convenient installation using migration functions Guaranteed consistency based on triggers, foreign-key constraints, and ACID No mutation without recorded Change No Change without Transaction Customizable audit metadata (per transaction) Clear query interfaces to the audit trail Optional processing &amp; purging logic following the Outbox pattern Based on Ecto and Postgrex with no further dependencies. No configuration of PostgreSQL database needed","ref":"readme.html#features","title":"README - Features","type":"extras"},{"doc":"Carbonite keeps a central changes table where all mutations of participating tables are recorded. On each such table a trigger is installed (after INSERT , UPDATE , and DELETE statements - TRUNCATE is not supported, see below) which calls a procedure stored within the database. This procedure captures the new or updated data within the changes table automatically. The procedure fetches its own per-table settings from another table called triggers . These settings customize the procedure's behaviour, for instance, they may exclude certain table columns from being captured. Besides the changes table, storing information about individual database statements (or rather, their impact on the data), Carbonite's transactions table stores information on database transactions . Within your application logic, you can insert a row into this table after you begin a database transaction, and record arbitrary metadata for it. Apart from the metadata, the transactions table houses two identifier columns: id and xact_id . While id is an ordinary, autoincrementing integer primary key, the xact_id is set to pg_current_xact_id() , PostgreSQL's internal transaction identifier. Each row in the changes table is associated to a single row in the transactions table, referencing both these identifiers . The changes.transaction_id (referencing transactions.id ) field is filled with the &quot;current value&quot; of the related sequence. This is your regular foreign key column, enough to relate a changes record with exactly one transactions record. Consistency with the transactions table is guaranteed by a foreign key constraint. Additionally, the changes.transaction_xact_id field is set to pg_current_xact_id() . Together these references ensure that, not only relates each changes record to a single transactions record, but records in these tables have been inserted within the same database transaction . Consistency is ensure by a manual lookup in the trigger procedure. This leads to the following interesting properties: All changes created within a database transaction automatically and implicitly belong to the same record in the transactions table, even if they're created separately and agnostic of each other in the application logic. This gives the developer a &quot;natural&quot; way to group related changes into events (more on events later). The entry in the transactions table must be created before any changes . Attempting to modify an audited table without prior insertion into the transactions table will result in an error. ℹ️ &ensp;Trigger vs. Write-Ahead-Log / Logical Decoding / Extensions Existing solutions for CDC on top of a PostgreSQL database (e.g. Debezium ) often tail the Write-Ahead-Log or use extensions (e.g. pgaudit ) instead of using triggers to create change records. Both approaches have their own advantages and disadvantages, and one should try to understand the differences before making up their mind. One of the main selling points of triggers is that they are executed as part of the normal transaction logic in the database, hence they benefit from the usual atomicity and consistency guarantees of relational databases. For audit triggers, among other things this means that data mutations and their audit trail are committed as an atomic unit. At the same time, this property of triggers strongly couples the auditing logic to the business operation. For instance, if the audit trigger raises an error, the entire transaction is aborted. And vice versa, if the business operation aborts, nothing is audited. Additionally, auditing unavoidably has an impact on the database performance. A common criticism of audit triggers is the performance penalty they incur on otherwise &quot;simple&quot; database operations, which is said to be much higher than just tailing the WAL. Carbonite tries to limit the work done in its trigger, but a few indexes and tables are nonetheless touched. Another advantage of triggers is their universality: You should be able to run Carbonite's migrations on any hosted PostgreSQL instance without the need to tweak its configuration or install custom extensions before. The primary advantage of triggers for Carbonite, though, is that we can make use of PostgreSQL's internal transaction id: It allows us to group related changes together and record metadata for the group. This and the simple tooling around it is not (as easily) achievable with a fully decoupled system. However, if you do not need this feature of Carbonite, please be sure to consider other solutions as well. ℹ️ &ensp;Why TRUNCATE is not supported As the last of the 4 primary data mutating SQL commands, TRUNCATE allows to delete all data from a table or a set of tables. It can be instrumented using triggers on the statement level ( FOR EACH STATEMENT instead of FOR EACH ROW ), which means the trigger procedure executes only once for the statement and without any data - in contrast to, for instance, an UPDATE statement, which also mutates multiple rows but fires the procedure once for each row. While there might be value in auditing TRUNCATE statements, the behaviour of the trigger procedure would be quite different from the other commands. For a rarely used SQL command, we chose against this additional complexity and not to support TRUNCATE in Carbonite.","ref":"readme.html#how-it-works","title":"README - How it works","type":"extras"},{"doc":"Requirements Due to its use of pg_current_xact_id , Carbonite requires PostgreSQL version 13 or above . If you see an error message like the following, your PostgreSQL installation is too old: ** (Postgrex.Error) ERROR 42704 (undefined_object) type &quot;xid8&quot; does not exist Hex dependency # mix.exs def deps do [ { :carbonite , &quot;~&gt; 0.6.0&quot; } ] end","ref":"readme.html#installation","title":"README - Installation","type":"extras"},{"doc":"As much of Carbonite's logic lives in database functions and triggers, to get started, we first need to create a migration using Ecto. Creating the initial migration Carbonite contains a Mix task that generates the initial migration for you. Please open the generated file and edit it according to your needs. See Carbonite.Migrations for more information on migrations. mix carbonite.gen.initial_migration -r MyApp.Repo The final migration should look something like this: # priv/repo/migrations/20210704201534_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do use Ecto.Migration def up do Carbonite.Migrations . up ( 1 ) Carbonite.Migrations . up ( 2 ) Carbonite.Migrations . up ( 3 ) Carbonite.Migrations . up ( 4 ) Carbonite.Migrations . up ( 5 ) # For each table that you want to capture changes of, you need to install the trigger. Carbonite.Migrations . create_trigger ( :rabbits ) # Optionally you may configure the trigger inside the migration. Carbonite.Migrations . put_trigger_config ( :rabbits , :excluded_columns , [ &quot;age&quot; ] ) # Optional outbox to process transactions later. Carbonite.Migrations . create_outbox ( &quot;rabbit_holes&quot; ) end def down do # Remove outbox again. Carbonite.Migrations . drop_outbox ( &quot;rabbit_holes&quot; ) # Remove all triggers before dropping the schema. Carbonite.Migrations . drop_trigger ( :rabbits ) # Drop the Carbonite tables. Carbonite.Migrations . down ( 5 ) Carbonite.Migrations . down ( 4 ) Carbonite.Migrations . down ( 3 ) Carbonite.Migrations . down ( 2 ) Carbonite.Migrations . down ( 1 ) end end Updates When a new Carbonite version is released, it may contain updates to the database schema. As these are announced in the Changelog , you each time need to create a migration in your host application like the following. # priv/repo/migrations/20210704201534_update_carbonite.exs defmodule MyApp.Repo.Migrations.UpdateCarbonite do use Ecto.Migration def up do Carbonite.Migrations . up ( 3 ) end def down do Carbonite.Migrations . down ( 3 ) end end Applying multiple Carbonite migrations in a single host migration is fine. Note that for each of your Carbonite &quot;partitions&quot; (see below), you need to run each Carbonite migration. Trigger configuration The behaviour of the capture trigger is customizable per table by manipulating the settings in the triggers table. Often you will want to update these settings within a migration, as well, which is why Carbonite provides the small helper function Carbonite.Migrations.put_trigger_config/4 that updates the settings using plain SQL statements. Primary Key Columns To speed up version lookups for a specific record, Carbonite copies its primary key(s) to the table_pk column of the changes table. The table keeps an index on this column together with the table prefix and name. By default, Carbonite will try to copy the :id column of the source table. If your table does not have a primary key, has a primary key with a different name, or has a composite primary key, you can override this using the primary_key_columns option. # Disable PK copying Carbonite.Migrations . put_trigger_config ( :rabbits , :primary_key_columns , [ ] ) # Different name Carbonite.Migrations . put_trigger_config ( :rabbits , :primary_key_columns , [ &quot;identifier&quot; ] ) # Composite PK Carbonite.Migrations . put_trigger_config ( :rabbits , :primary_key_columns , [ &quot;house&quot; , &quot;apartment_no&quot; ] ) Since the changes table keeps versions of a multitude of different source tables, primary keys are first cast to string (the table_pk column has type VARCHAR[] ). For composite primary keys, set the primary_key_columns option to an array as shown above. Each component of a compound primary key will be cast to string before the components are joined into the array. Excluded and Filtered Columns In case your table contains sensitive data or data otherwise undesirable for change capturing, you can exclude columns using the excluded_columns option. Excluded columns will not appear in the captured data. If an UPDATE on a table solely touches excluded columns, the entire UPDATE will not be recorded. Carbonite.Migrations . put_trigger_config ( :rabbits , :excluded_columns , [ &quot;age&quot; ] ) If you still want to capture changes to a column (in the changed field), but don't need the exact data, you can make it a &quot;filtered&quot; column. These columns appear as [FILTERED] in the data field. Carbonite.Migrations . put_trigger_config ( :rabbits , :filtered_columns , [ &quot;age&quot; ] ) Persisting replaced data on UPDATE statements By default, the changes table will track the current version of a record in its data field and a list of changed fields in the changed column. Optionally, you can decide to store the replaced data (the &quot;diff&quot;) in the changed_from field, by enabling the store_changed_from option for a trigger. This defaults to false . Carbonite.Migrations . put_trigger_config ( :rabbits , :store_changed_from , true ) Partitioning the Audit Trail Carbonite can install its tables into multiple database schemas using the carbonite_prefix option. You can use this feature to &quot;partition&quot; your captured data. Carbonite.Migrations . up ( 1 , carbonite_prefix : &quot;carbonite_animals&quot; ) Carbonite.Migrations . create_trigger ( :rabbits , carbonite_prefix : &quot;carbonite_animals&quot; ) Basically all of Carbonite's functions accept the same carbonite_prefix option to target a particular partition. If desired, tables can participate in multiple partitions by adding multiple triggers on to them. Keep in mind that each partition will need to be processed and purged separately, resulting in multiple streams of change data in your external storage.","ref":"readme.html#getting-started","title":"README - Getting started","type":"extras"},{"doc":"In your application logic, before modifying an audited table like rabbits , you need to first create a Carbonite.Transaction record. With Ecto.Multi The easiest way to do so is using Carbonite.Multi.insert_transaction/2 within an Ecto.Multi operation: Ecto.Multi . new ( ) |&gt; Carbonite.Multi . insert_transaction ( %{ meta : %{ type : &quot;rabbit_inserted&quot; } } ) |&gt; Ecto.Multi . insert ( :rabbit , &amp; MyApp.Rabbit . create_changeset ( &amp;1 . params ) ) |&gt; MyApp.Repo . transaction ( ) As you can see, the Carbonite.Transaction is a great place to store metadata for the operation. A type field can be used to categorize the transactions. A user_id would be a good candidate for a transaction log, as well. Building a changeset for manual insertion If you don't have the luxury of an Ecto.Multi , you can create a changeset for a Carbonite.Transaction using Carbonite.Transaction.changeset/1 : MyApp.Repo . transaction ( fn -&gt; %{ meta : %{ type : &quot;rabbit_inserted&quot; } } |&gt; Carbonite.Transaction . changeset ( ) |&gt; MyApp.Repo . insert! ( ) # ... end ) Setting metadata outside of the transaction In case you do not have access to metadata you want to persist in the Carbonite.Transaction at the code site where you create it, you can use Carbonite.Transaction.put_meta/2 to store metadata in the process dictionary . This metadata is merged into the metadata given to Carbonite.Multi.insert_transaction/2 . # e.g., in a controller or plug Carbonite.Transaction . put_meta ( :user_id , ... ) Transactions in data migrations If you manipulate data inside your transactions, as usual a Carbonite.Transaction needs to be inserted before any other statements. You can use Carbonite.Migrations.insert_migration_transaction/1 to insert a transaction with a meta attribute populated from the migration module. import Carbonite.Migrations def change do insert_migration_transaction ( ) execute ( &quot;UPDATE ...&quot; ) end","ref":"readme.html#inserting-a-transaction","title":"README - Inserting a Transaction","type":"extras"},{"doc":"Of course, persisting the audit trail is not an end in itself. At some point you will want to read the data back and make it accessible to the user. Carbonite.Query offers a small suite of helper functions that make it easier to query the database for Transaction and Change records. Fetching transactions The Carbonite.Query.transactions/1 function constructs an Ecto.Query for loading Carbonite.Transaction.t/0 records from the database, optionally preloading their included changes. The query can be further refined to limit the result set. Carbonite.Query . transactions ( ) |&gt; Ecto.Query . where ( [ t ] , t . inserted_at &gt; ^ earliest ) |&gt; MyApp.Repo . all ( ) Fetching changes of invididual records The Carbonite.Query.changes/2 function constructs an Ecto.Query.t/0 from a schema struct, loading all changes stored for the given source record. % MyApp.Rabbit { id : 1 } |&gt; Carbonite.Query . changes ( ) |&gt; MyApp.Repo . all ( )","ref":"readme.html#retrieving-data","title":"README - Retrieving data","type":"extras"},{"doc":"Storing event series in a relational database comes with the usual scaling issues, so at some point you may find it necessary to offload the captured data to an external data store. For processing, exporting, and later purging the captured transactions, Carbonite includes what is called an &quot;Event Outbox&quot;, or in fact as many of those as you need. Essentially, a Carbonite.Outbox is a named cursor on the ordered list of transactions stored in an audit trail. Each outbox advances by calling a user-supplied processor function on a batch of transactions. The batch can be filtered and limited in size, but is always ordered ascending by the transaction's id attribute, so transactions are processed roughly in order of their insertion (see below for a discussion on caveats of this solution). When a batch is processed, the outbox remembers its last position - the last transaction that has been processed - so the following call to the outbox can continue where the last has left of. Besides, the outbox persists a user-definable &quot;memo&quot;, an arbitrary map that allows to feed output of the previous processor run into the next one. Creating an Outbox Create an outbox within a migration: Carbonite.Migrations . create_outbox ( &quot;rabbit_holes&quot; ) You can create more than one outbox if you need multiple processors, e.g. for multiple target systems for your data. Processing You can process an outbox by calling Carbonite.process/4 with the outbox name and a processor callback function. Carbonite . process ( MyApp.Repo , &quot;rabbit_holes&quot; , fn transactions , _memo -&gt; for transaction &lt;- transactions do send_to_external_database ( transaction ) end :cont end ) In practise, you will almost always want to run this within an asynchronous job processor. The following shows an exemplary Oban worker: # config/config.exs config :my_app , :oban , repo : MyApp.Repo , plugins : [ { Oban.Plugins.Cron , crontab : [ { &quot;0 0 * * *&quot; , MyApp.PeriodicOutboxWorker , args : %{ outbox : &quot;rabbit_holes&quot; } } ] ] # lib/my_app/periodic_outbox_worker.ex defmodule MyApp.PeriodicOutboxWorker do # Worker is scheduled every 24h, let&#39;s make it unique for 12h to be sure. # Retry is disabled, the next scheduled run will pick up failed outboxes. use Oban.Worker , queue : :default , unique : [ period : 43_200 ] , retry : false def perform ( % Oban.Job { args : %{ &quot;outbox&quot; =&gt; outbox } } ) do Carbonite . process ( MyApp.Repo , outbox , fn transactions , _memo -&gt; for transaction &lt;- transactions do send_to_external_database ( transaction ) end :cont end ) end end ⚠️&ensp;Transactionality & mutual exclusion As the outbox processor is likely to call external services as part of its work, Carbonite.process/4 does not begin a transaction itself, and consequently does not acquire a lock on the outbox record. In other words, users have to ensure that they don't accidentally run Carbonite.process/4 for the same outbox concurrently, for instance making use of uniqueness options of the used job processor (e.g. unique in Oban ). The absence of transactionality also means that any exception raised within the processor function will immediately abort the Carbonite.process/4 call without writing the current outbox position to disk. As a result, transactions may be processed again in the next run. Please make sure your receiving system or database can handle duplicate messages. ⚠️&ensp;Long running / parallel transactions and the outbox order When a Carbonite.Transaction record is created at the beginning of an operation in your application, it records the &quot;current&quot; sequence value in its id field. The record in the transactions table becomes visible to other transactions when the current transaction is committed. A few observations can be made from this: When ordered by their id field, the transactions will be roughly sorted by the time the corresponding operation was started, not when it was committed. Two transactions running in parallel may be committed &quot;out of order&quot;, i.e. one with the larger id may be committed before the smaller id transaction if that has a longer runtime. The latter point is crucial: For instance, if two transactions with id=1 and id=2 run in parallel and id=2 finishes before id=1 , an outside viewer can already see id=2 in the database before id=1 is committed. If this outside viewer happens to be an outbox processing job, transaction with id=1 might be skipped and never looked at again. To mitigate this issue, Carbonite.process/4 has a min_age option which excludes transaction younger than a certain from the processing batch (5 minutes by default, increase this is you expect longer running transactions). Purging Now that you have successfully exported your audit data, you may delete old transactions from the primary database. Carbonite.purge/2 deletes records that have been processed by all existing outboxes. # Deletes records whose id field is less than the transaction_id on each outbox. Carbonite . purge ( MyApp.Repo )","ref":"readme.html#processing-data","title":"README - Processing data","type":"extras"},{"doc":"One of Carbonite's key features is that it is virtually impossible to forget to record a change to a table (due to the trigger) or to forget to insert an enclosing Carbonite.Transaction beforehand (due to the foreign key constraint between changes and transactions ). However, in some circumstances it may be desirable to temporarily switch off change capturing. One such situation is the use of factories (e.g. ExMachina) inside your test suite: Inserting a transaction before each factory call quickly becomes cumbersome and will unnecessarily increase execution time. To bypass the capture trigger, Carbonite's trigger configuration provides a toggle mechanism consisting of two fields: mode and override_xact_id . The former you set while installing the trigger on a table in a migration, while the latter allows to &quot;override&quot; whatever has been set at runtime, and only for the current transaction. If you are using Ecto's SQL sandbox for running transactional tests, this means the override is going to be active until the end of the test case. As a result, you have two options: Leave the mode at the default value of :capture and turn off capturing as needed by switching to &quot;override mode&quot;. This means for every test case where you do not care about change capturing, you explicitly disable the trigger before any database calls; for instance, in an ExUnit setup block. This approach has the benefit that you still capture all changes by default, and can't miss to test a code path that (in production) would require a Carbonite.Transaction . It is, however, still pretty expensive at ~1 additional SQL call per test case. Set the mode to :ignore on all triggers in your :test environment and instead selectively turn on capturing in test cases where you want to assert on the captured data. For instance, you can set the trigger mode in your migration based on the Mix environment. This approach is cheaper as it does not require any action in your tests by default. Yet you should make sure that you test all code paths that do mutate change-captured tables, in order to assert that each of these inserts a transaction as well. The following code snippet illustrates the second approach: # config/config.exs config :my_app , carbonite_mode : :capture # config/test.exs config :my_app , carbonite_mode : :ignore # priv/repo/migrations/000000000000_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do @mode Application . compile_env! ( :my_app , :carbonite_mode ) def up do # ... Carbonite.Migrations . create_trigger ( :rabbits ) Carbonite.Migrations . put_trigger_config ( :rabbits , :mode , @mode ) end end # test/support/carbonite_helpers.exs defmodule MyApp.CarboniteHelpers do def carbonite_override_mode ( _ ) do Carbonite . override_mode ( MyApp.Repo ) :ok end def current_transaction_meta do Carbonite.Query . current_transaction ( ) |&gt; MyApp.Repo . one! ( ) |&gt; Map . fetch ( :meta ) end end # test/some_test.exs describe &quot;my_operation/0&quot; do setup [ :carbonite_override_mode ] test &quot;auditing&quot; do my_operation ( ) assert current_transaction_meta ( ) == %{ &quot;type&quot; =&gt; &quot;some_operation&quot; } end end","ref":"readme.html#testing-bypassing-carbonite","title":"README - Testing / Bypassing Carbonite","type":"extras"},{"doc":"Inspiration The trigger-based table versioning derives from audit_trigger_91plus , an &quot;example of a generic trigger function&quot; hosted in the PostgreSQL wiki. Artwork The amazing Carbonite logo has been designed by Petra Herberger . Thank you! 💜","ref":"readme.html#acknowledgements","title":"README - Acknowledgements","type":"extras"},{"doc":"Copyright © 2021–2022, Bitcrowd GmbH. You may use and redistribute Carbonite and its source code under the Apache 2.0 License, see the LICENSE file for details.","ref":"readme.html#copyright-and-license","title":"README - Copyright and License","type":"extras"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"","ref":"changelog.html#unreleased","title":"Changelog - Unreleased","type":"extras"},{"doc":"New migration patches: 5 Added Optional tracking of previous data in changes. Set the store_changed_from trigger option. Changed Added @schema_prefix &quot;carbonite_default&quot; on all schemas. This will enable the manual usage of Repo.insert/2 &amp; friends on transactions without specifying a prefix, when using the default carbonite prefix.","ref":"changelog.html#0-6-0-2022-10-12","title":"Changelog - [0.6.0] - 2022-10-12","type":"extras"},{"doc":"New migration patches: 4 Switch to normal identity column on transactions The id column on transactions has been replaced with an ordinary autoincrementing integer PK, filled from a sequence. Next to it a new xact_id column continues to store the transaction id (from pg_current_xact_id ). Both values used together ensure that, first the id is monotonically increasing and survives a backup restore (see issue #45), and second the changes records can still only be inserted within the same transaction. Added Carbonite.Migrations.insert_migration_transaction/1 and its macro friend, insert_migration_transaction_after_begin , help with data migrations. Carbonite.fetch_changes returns the changes of the current (ongoing) transaction.","ref":"changelog.html#0-5-0-2022-02-25","title":"Changelog - [0.5.0] - 2022-02-25","type":"extras"},{"doc":"New migration patches: 2, 3 Switch to top-level API with repo param Carbonite.override_mode/2 (kept a wrapper in Carbonite.Multi ) Carbonite.insert_transaction/3 (kept a wrapper in Carbonite.Multi ) Carbonite.process/4 (previously Carbonite.Outbox.process/3 with major changes) Carbonite.purge/2 (previously Carbonite.Outbox.purge/1 with major changes) Big outbox overhaul Split into query / processing Simplify processing No more transaction New capabilities: memo, halting, chunking Migration versioning Explicit for now with Carbonite.Migrations.up(non_neg_integer()) Carbonite.Migrations.install_schema/1 is now Carbonite.Migrations.up/2 Carbonite.Migrations.put_trigger_option/4 to ensure old migrations continue to work At the same time removed long configuration statement from Carbonite.Migrations.install_trigger/1 , so this does not need to be versioned and continues to work Mix task for generating the &quot;initial&quot; migration Other Changes Optionally derive Jason.Encoder for Carbonite.Transaction and Carbonite.Change Made all prefix options binary-only (no atom) as Ecto.Query.put_query_prefix/2 only accepts strings","ref":"changelog.html#0-4-0-2021-11-07","title":"Changelog - [0.4.0] - 2021-11-07","type":"extras"},{"doc":"Added table_prefix option to Query.changes/2 allows to override schema prefix of given record Query.transactions/1 query selects all transactions","ref":"changelog.html#0-3-1-2021-10-23","title":"Changelog - [0.3.1] - 2021-10-23","type":"extras"},{"doc":"Added Carbonite.Migrations.drop_tables/1 allows to drop the carbonite audit trail without removing the schema Changed Renamed the option that can be passed to Carbonite.Migrations.drop_schema/1 from prefix to carbonite_prefix Changed Carbonite.Migrations.drop_schema/1 to also drop the tables Made Carbonite.Multi.insert_transaction/3 ignore conflicting INSERT s within the same transaction Also, changed Carbonite.Multi.insert_transaction/3 to always reloads all fields from the database after insertion, immediately returning the JSONinified meta payload Fixed Fixed ignore mode when override_transaction_id is NULL","ref":"changelog.html#0-3-0-2021-10-22","title":"Changelog - [0.3.0] - 2021-10-22","type":"extras"},{"doc":"Fixed Fixed broken documentation","ref":"changelog.html#0-2-1-2021-10-10","title":"Changelog - [0.2.1] - 2021-10-10","type":"extras"},{"doc":"Added Store primary key(s) on changes table and index them Add Carbonite.Query module current_transaction/2 allows to fetch the ongoing transaction (for sandbox tests) changes/2 allows to fetch the changes of an invidual source record Update Postgrex to 0.15.11 and drop local Xid8 type Add mode field to trigger (capture or ignore) Add &quot;override mode&quot; reversing the mode option for the current transaction to enable/disable capturing on demand (e.g. in tests) Add filtered columns Changed Moved top-level functions to nested modules Transaction and Multi Made table_pk be NULL when primary_key_columns is an empty array Default primary_key_columns to [&quot;id&quot;] Renamed prefix option to carbonite_prefix on install_schema/2 for consistency","ref":"changelog.html#0-2-0-2021-10-10","title":"Changelog - [0.2.0] - 2021-10-10","type":"extras"},{"doc":"Initial release.","ref":"changelog.html#0-1-0-2021-09-01","title":"Changelog - [0.1.0] - 2021-09-01","type":"extras"}]